import multiprocessing
import os
import sys

from django.apps import apps
from django.core.management.base import BaseCommand
from django.utils.module_loading import module_has_submodule

#: The number of available CPUs.
CPU_COUNT = multiprocessing.cpu_count()


class Command(BaseCommand):
    help = "Runs Dramatiq workers."

    def add_arguments(self, parser):
        parser.add_argument(
            "--no-reload",
            action="store_false",
            dest="use_watcher",
            help="Disable autoreload.",
        )
        parser.add_argument(
            "--reload-use-polling",
            action="store_true",
            dest="use_polling_watcher",
            help=(
                "Use a poll-based file watcher for autoreload (useful under "
                "Vagrant and Docker for Mac)."
            ),
        )
        parser.add_argument(
            "--use-gevent",
            action="store_true",
            help="Use gevent for worker concurrency.",
        )
        parser.add_argument(
            "--processes", "-p",
            default=CPU_COUNT,
            type=int,
            help="The number of processes to run (default: %d)." % CPU_COUNT,
        )
        parser.add_argument(
            "--threads", "-t",
            default=CPU_COUNT,
            type=int,
            help="The number of threads per process to use (default: %d)." % CPU_COUNT,
        )
        parser.add_argument(
            "--path", "-P",
            default=".",
            nargs="*",
            type=str,
            help="The import path (default: .).",
        )

    def handle(self, use_watcher, use_polling_watcher, use_gevent, path, processes, threads, verbosity, **options):
        executable_name = "dramatiq-gevent" if use_gevent else "dramatiq"
        executable_path = self._resolve_executable(executable_name)
        watch_args = ["--watch", "."] if use_watcher else []
        if watch_args and use_polling_watcher:
            watch_args.append("--watch-use-polling")

        verbosity_args = ["-v"] * (verbosity - 1)
        tasks_modules = self.discover_tasks_modules()
        process_args = [
            executable_name,
            "--path", *path,
            "--processes", str(processes),
            "--threads", str(threads),

            # --watch /path/to/project [--watch-use-polling]
            *watch_args,

            # -v -v ...
            *verbosity_args,

            # django_dramatiq.tasks app1.tasks app2.tasks ...
            *tasks_modules,
        ]

        self.stdout.write(' * Running dramatiq: "%s"\n\n' % " ".join(process_args))
        os.execvp(executable_path, process_args)

    def discover_tasks_modules(self):
        app_configs = apps.get_app_configs()
        tasks_modules = ["django_dramatiq.setup"]
        for conf in app_configs:
            if module_has_submodule(conf.module, "tasks"):
                module = conf.name + ".tasks"
                tasks_modules.append(module)
                self.stdout.write(" * Discovered tasks module: %r" % module)
        return tasks_modules

    def _resolve_executable(self, exec_name):
        bin_dir = os.path.dirname(sys.executable)
        if bin_dir:
            return os.path.join(bin_dir, exec_name)
        return exec_name



import dramatiq


@dramatiq.actor
def delete_old_tasks(max_task_age=86400):
    """This task deletes all tasks older than `max_task_age` from the
    database.
    """
    from .models import Task
    Task.tasks.delete_old_tasks(max_task_age)



import logging

from django import db
from dramatiq.middleware import Middleware

from .models import Task

LOGGER = logging.getLogger("django_dramatiq.AdminMiddleware")


class AdminMiddleware(Middleware):
    """This middleware keeps track of task executions.
    """

    def after_enqueue(self, broker, message, delay):
        LOGGER.debug("Creating Task from message %r.", message.message_id)
        status = Task.STATUS_ENQUEUED
        if delay:
            status = Task.STATUS_DELAYED

        Task.tasks.create_or_update_from_message(message, status=status)

    def before_process_message(self, broker, message):
        LOGGER.debug("Updating Task from message %r.", message.message_id)
        Task.tasks.create_or_update_from_message(message, status=Task.STATUS_RUNNING)

    def after_process_message(self, broker, message, *, result=None, exception=None):
        status = Task.STATUS_DONE
        if exception is not None:
            status = Task.STATUS_FAILED

        LOGGER.debug("Updating Task from message %r.", message.message_id)
        Task.tasks.create_or_update_from_message(message, status=status)


class DbConnectionsMiddleware(Middleware):
    """This middleware cleans up db connections on worker shutdown.
    """

    def _close_connections(self, *args, **kwargs):
        db.connections.close_all()

    before_consumer_thread_shutdown = _close_connections
    before_worker_thread_shutdown = _close_connections
    before_worker_shutdown = _close_connections



import json

from datetime import datetime
from django.contrib import admin
from django.utils import timezone
from django.utils.safestring import mark_safe

from .models import Task


@admin.register(Task)
class TaskAdmin(admin.ModelAdmin):
    exclude = ("message_data",)
    readonly_fields = ("message_details", "traceback", "status")
    list_display = ("__str__", "status", "eta", "created_at", "updated_at", "queue_name", "actor_name")

    def queue_name(self, instance):
        return instance.message.queue_name

    def actor_name(self, instance):
        return instance.message.actor_name

    def eta(self, instance):
        timestamp = instance.message.options.get("eta", instance.message.message_timestamp) / 1000
        return timezone.make_aware(datetime.utcfromtimestamp(timestamp))

    def message_details(self, instance):
        message_details = json.dumps(instance.message._asdict(), indent=4)
        return mark_safe("<pre>%s</pre>" % message_details)

    def traceback(self, instance):
        traceback = instance.message.options.get("traceback", None)
        if traceback:
            return mark_safe("<pre>%s</pre>" % traceback)
        return None



import importlib


def load_class(path):
    try:
        module_path, _, class_name = path.rpartition(".")
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except AttributeError:
        raise ImportError("Module '%s' doesn't have a class named '%s'." % (
            module_path, class_name,
        ))


def load_middleware(path_or_obj):
    return load_class(path_or_obj)()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-10-31 06:58
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.manager


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.UUIDField(editable=False, primary_key=True, serialize=False)),
                ('status', models.CharField(choices=[('enqueued', 'Enqueued'), ('delayed', 'Delayed'), ('running', 'Running'), ('failed', 'Failed'), ('done', 'Done')], default='enqueued', max_length=8)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('message_data', models.BinaryField()),
            ],
            options={
                'ordering': ['-updated_at'],
            },
            managers=[
                ('tasks', django.db.models.manager.Manager()),
            ],
        ),
    ]



from datetime import timedelta
from django.db import models, utils
from django.utils.functional import cached_property
from django.utils.timezone import now
from dramatiq import Message

from .apps import DjangoDramatiqConfig

#: The database label to use when storing task metadata.
DATABASE_LABEL = DjangoDramatiqConfig.tasks_database()


class TaskManager(models.Manager):
    def create_or_update_from_message(self, message, **extra_fields):
        try:
            return self.using(DATABASE_LABEL).create(
                id=message.message_id,
                message_data=message.encode(),
                **extra_fields,
            )

        except (utils.OperationalError, utils.IntegrityError):
            task = self.using(DATABASE_LABEL).get(id=message.message_id)
            task.message_data = message.encode()
            for name, value in extra_fields.items():
                setattr(task, name, value)

            task.save()
            return task

    def delete_old_tasks(self, max_task_age):
        self.using(DATABASE_LABEL).filter(
            created_at__lte=now() - timedelta(seconds=max_task_age)
        ).delete()


class Task(models.Model):
    STATUS_ENQUEUED = "enqueued"
    STATUS_DELAYED = "delayed"
    STATUS_RUNNING = "running"
    STATUS_FAILED = "failed"
    STATUS_DONE = "done"
    STATUSES = [
        (STATUS_ENQUEUED, "Enqueued"),
        (STATUS_DELAYED, "Delayed"),
        (STATUS_RUNNING, "Running"),
        (STATUS_FAILED, "Failed"),
        (STATUS_DONE, "Done"),
    ]

    id = models.UUIDField(primary_key=True, editable=False)
    status = models.CharField(max_length=8, choices=STATUSES, default=STATUS_ENQUEUED)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    message_data = models.BinaryField()

    tasks = TaskManager()

    class Meta:
        ordering = ["-updated_at"]

    @cached_property
    def message(self):
        return Message.decode(bytes(self.message_data))

    def __str__(self):
        return str(self.message)



import dramatiq

from django.apps import AppConfig
from django.conf import settings

from .utils import load_class, load_middleware

DEFAULT_BROKER = "dramatiq.brokers.rabbitmq.RabbitmqBroker"
DEFAULT_SETTINGS = {
    "BROKER": DEFAULT_BROKER,
    "OPTIONS": {
        "host": "127.0.0.1",
        "port": 5672,
        "heartbeat_interval": 0,
        "connection_attempts": 5,
    },
    "MIDDLEWARE": [
        "dramatiq.middleware.Prometheus",
        "dramatiq.middleware.AgeLimit",
        "dramatiq.middleware.TimeLimit",
        "dramatiq.middleware.Callbacks",
        "dramatiq.middleware.Retries",
        "django_dramatiq.middleware.AdminMiddleware",
        "django_dramatiq.middleware.DbConnectionsMiddleware",
    ]
}


class DjangoDramatiqConfig(AppConfig):
    name = "django_dramatiq"
    verbose_name = "Django Dramatiq"

    def ready(self):
        broker_settings = type(self).broker_settings()
        broker_path = broker_settings["BROKER"]
        broker_class = load_class(broker_path)
        broker_options = broker_settings.get("OPTIONS", {})
        middleware = [load_middleware(path) for path in broker_settings.get("MIDDLEWARE", [])]
        broker = broker_class(middleware=middleware, **broker_options)
        dramatiq.set_broker(broker)

    @classmethod
    def broker_settings(cls):
        return getattr(settings, "DRAMATIQ_BROKER", DEFAULT_SETTINGS)

    @classmethod
    def tasks_database(cls):
        return getattr(settings, "DRAMATIQ_TASKS_DATABASE", "default")
